#Region Header

#cs

	Title:			Management of Hotkeys UDF Library for AutoIt3
	Filename:		HotKey.au3
	Description:	Sets a hotkey that calls a user function
	Author:			Yashied
	Version:		1.4
	Requirements:	AutoIt v3.3 +, Developed/Tested on WindowsXP Pro Service Pack 2
	Uses:			StructureConstants.au3, WinAPI.au3, WindowsConstants.au3
	Notes:			-

	Available functions:
	
	_HotKeyAssign
	_HotKeyEnable
	_HotKeyDisable
	_HotKeyRelease

	Example1:

		#Include <HotKey.au3>

		const $VK_ESCAPE = 0x1B
		const $VK_F12 = 0x7B

		_HotKeyAssign(BitOR($CK_CONTROL, $VK_F12), 'Message'); Assign CTRL-F12 with Message()
		_HotKeyAssign($VK_ESCAPE, 'Quit'); Assign ESC with Quit()

		while 1
			Sleep(10)
		wend

		func Message()
			MsgBox(0, 'Hotkey Test Message', 'CTRL-F12 has been pressed!')
		endfunc; Message

		func Quit()
			exit
		endfunc; Quit

	Example2:

		#Include <GUIConstantsEx.au3>
		#Include <HotKey.au3>

		const $VK_F12 = 0x7B

		local $Form, $Label
		local $i = 0

		$Form = GUICreate('Test', 200, 200)
		$Label = GUICtrlCreateLabel($i, 20, 72, 160, 52, 0x01)
		GUICtrlSetFont(-1, 32, 400, 0, 'Tahoma')
		GUISetState()
		
		_HotKeyAssign(BitOR($CK_CONTROL, $VK_F12), 'MyFunc', $Form, 0, 500, 50); Assign CTRL-F12 with MyFunc() for created window and define repeat delays

		do
		until GUIGetMsg() = $GUI_EVENT_CLOSE

		func MyFunc()
			$i += 1
			GUICtrlSetData($Label, $i)
		endfunc; MyFunc

#ce

#Include-once

#Include <StructureConstants.au3>
#Include <WinAPI.au3>
#Include <WindowsConstants.au3>

#EndRegion Header

#Region Global Variables and Constants

global const $CK_SHIFT = 0x0100
global const $CK_CONTROL = 0x0200
global const $CK_ALT = 0x0400
global const $CK_WIN = 0x0800

#EndRegion Global Variables and Constants

#Region Local Variables and Constants

dim $hkTb[1][9] = [[0, 0, 0, 1, 0]]

#cs

DO NOT USE THIS ARRAY IN THE SCRIPT, INTERNAL USE ONLY!

$hkTb[0][0]   - Count item of array
	 [0][1]   - Interruption control flag (need to set this flag before changing $hkTb array)
	 [0][2]   - Last key pressed (16-bit code)
	 [0][3]   - Allow the hotkeys control flag (see _HotKeyEnable() and _HotKeyDisable())
     [0][4]   - Handle to the hook procedure
     [0][5-8] - Don`t used
	 
$hkTb[i][0]   - Combined hotkey code (see _HotKeyAssign)
	 [i][1]   - User function name
	 [i][2]   - Repeat delay, ms
	 [i][3]   - Repeat rate, ms
	 [i][4]   - TimeStamp for repeat delay (return by TimeInit())
	 [i][5]   - TimeStamp for repeat rate (return by TimeInit())
	 [i][6]   - Call next hook procedure control flag (see _HotKeyAssign())
	 [i][7]   - The title of the window to allow the hotkey
	 [i][8]   - Reserved

#ce

local $OnHotKeyExit = Opt('OnExitFunc', 'OnHotKeyExit')

#EndRegion Local Variables and Constants

#Region Public Functions

; #FUNCTION# ========================================================================================================================
; Function Name:	_HotKeyAssign
; Description:		Sets a hotkey that calls a user function.
; Syntax:			_HotKeyAssign ( $iKey, $sFunction [, $iTitle [, $iFlag [, $iDelay [, $iRate]]]] )
; Parameter(s):		$iKey      - Combined 16-bit hotkey code, which consists of upper and lower bytes. Value of bits shown in the following table.
;
;								 Hotkey code bits:
;
;								 0-7   - Specifies the virtual-key (VK) code of the key. Codes for the mouse buttons (0x01 - 0x06) are not supported.
;										 (http://msdn.microsoft.com/en-us/library/dd375731(VS.85).aspx)
;
;								 8     - SHIFT key
;								 9     - CONTROL key
;								 10    - ALT key
;								 11    - WIN key
;								 12-15 - Don`t used
;
;								 The combination of keys can be composed of one function key (VK) and some system keys (CK). This code can not consist
;								 only of the system keys. To combine system keys with the function key use BitOR(). See following examples.
;								 (VK-constants are used in the vkConstants.au3)
;
;								 SHIFT-F12    - BitOR($CK_SHIFT, $VK_F12)
;								 CTRL-A       - BitOR($CK_CONTROL, 0x41)
;								 CTRL-ALT-TAB - BitOR($CK_CONTROL, $CK_ALT, $VK_TAB)
;
;								 Not acceptable combinations:
;
;								 SHIFT-WIN    - BitOR($CK_SHIFT, $CK_WIN) - There is no function key.
;								 CTRL-Q-W     - BitOR($CK_CONTROL, 0x51, 0x57) - You can not specify more than one function key.
;					
;					$sFunction - The name of the function to call when the key is pressed. Not specifying this parameter will unset a previous hotkey.
;
;					$iTitle    - [optional] The title of the window to allow the hotkey. This parameter is the same as the WinActive() function.
;								 The default is 0, which is allow hotkey for all window.
;
;					$iFlag     - [optional] Control flag to lock the specified hot keys for other programs. Where $iFlag = 0 (default), then other programs
;								 can not use the specified hotkey. This may not work for certain system events, such as CTRL-ALT-DEL.
;
;					$iDelay    - [optional] Repeat delay (in milliseconds). Adjusts the amount of time that elapses before a character begins to repeat
;								 when you hold down a key. If $iDelay = 0 repeat are disabled. The default is 0, which is no repeat.
;
;					$iRate     - [optional] Repeat rate (in milliseconds). Adjusts the speed at which a character repeats when you hold down a key.
;								 If $iDelay = 0, the $iRate is ignored. The default is 50.
;
; Return Value(s):	Success: Returns 1.
;					Failure: Returns 0 and sets the @error flag to non-zero.
; Author(s):		Yashied
; Note(s):			-
;====================================================================================================================================

func _HotKeyAssign($iKey, $sFunction, $iTitle = 0, $iFlag = 0, $iDelay = 0, $iRate = 50)

	local $i, $j = 0, $k
	
	if BitAND($iKey, 0x00FF) = 0 then
		return SetError(1, 0, 0)
	endif
	
	if $iDelay < 0 then
		$iDelay = 0
	endif
	if $iRate < 0 then
		$iRate = 50
	endif
	$sFunction = StringStripWS($sFunction, 3)
	for $i = 1 to $hkTb[0][0]
		if $hkTb[$i][0] = $iKey then
			$j = $i
			exitloop
		endif
	next
	
	$hkTb[0][1] = 1
	
	if $j = 0 then
		if $sFunction = '' then
			$hkTb[0][1] = 0
			return SetError(0, 0, 1)
		endif
		if ($hkTb[0][0] = 0) and ($hkTb[0][3] = 1) and ($hkTb[0][4] = 0) then
			$hkTb[0][4] = _WinAPI_SetWindowsHookEx($WH_KEYBOARD_LL, DllCallbackGetPtr(DllCallbackRegister('_hook', 'long', 'int;wparam;lparam')), _WinAPI_GetModuleHandle(0), 0)
			if (@error) or ($hkTb[0][4] = 0) then
				$hkTb[0][1] = 0
				return SetError(1, 0, 0)
			endif
		endif
		$j = $hkTb[0][0] + 1
		redim $hkTb[$j + 1][9]
		$hkTb[0][0] += 1
	endif
	if $sFunction > '' then
		$hkTb[$j][0] = $iKey
		$hkTb[$j][1] = $sFunction
		$hkTb[$j][2] = $iDelay
		$hkTb[$j][3] = $iRate
		$hkTb[$j][6] = not ($iFlag = 0)
		$hkTb[$j][7] = $iTitle
		$hkTb[$j][8] = 0
	else
		for $i = $j to $hkTb[0][0] - 1
			for $k = 0 to 8
				$hkTb[$i][$k] = $hkTb[$i + 1][$k]
			next
		next
		redim $hkTb[$hkTb[0][0]][9]
		$hkTb[0][0] -= 1
		if $hkTb[0][0] = 0 then
			if _WinAPI_UnhookWindowsHookEx($hkTb[0][4]) then
				$hkTb[0][4] = 0
			endif
			$hkTb[0][2] = 0
		endif
	endif
	
	$hkTb[0][1] = 0
	
	return SetError(0, 0, 1)
endfunc; _HotKeyAssign

; #FUNCTION# ========================================================================================================================
; Function Name:	_HotKeyEnable
; Description:		Enables all the hot keys that were defined _HotKeyAssign() function.
; Syntax:			_HotKeyEnable (  )
; Parameter(s):		None.
; Return Value(s):	Success: Returns 1.
;					Failure: Returns 0 and sets the @error flag to non-zero.
; Author(s):		Yashied
; Note(s):			-
;====================================================================================================================================

func _HotKeyEnable()
	if ($hkTb[0][0] > 0) and ($hkTb[0][4] = 0) then
		$hkTb[0][4] = _WinAPI_SetWindowsHookEx($WH_KEYBOARD_LL, DllCallbackGetPtr(DllCallbackRegister('_hook', 'long', 'int;wparam;lparam')), _WinAPI_GetModuleHandle(0), 0)
		if (@error) or ($hkTb[0][4] = 0) then
			return SetError(1, 0, 0)
		endif
	endif
	$hkTb[0][3] = 1
	return SetError(0, 0, 1)
endfunc; _HotKeyEnable

; #FUNCTION# ========================================================================================================================
; Function Name:	_HotKeyDisable
; Description:		Disables all the hot keys that were defined _HotKeyAssign() function.
; Syntax:			_HotKeyDisable (  )
; Parameter(s):		None.
; Return Value(s):	Success: Returns 1.
;					Failure: Returns 0 and sets the @error flag to non-zero.
; Author(s):		Yashied
; Note(s):			-
;====================================================================================================================================

func _HotKeyDisable()
	if IsPtr($hkTb[0][4]) then
		if not _WinAPI_UnhookWindowsHookEx($hkTb[0][4]) then
			return SetError(1, 0, 0)
		endif
		$hkTb[0][2] = 0
		$hkTb[0][4] = 0
	endif
	$hkTb[0][3] = 0
	return SetError(0, 0, 1)
endfunc; _HotKeyDisable

; #FUNCTION# ========================================================================================================================
; Function Name:	_HotKeyRelease
; Description:		Removes all the hot keys that were defined _HotKeyAssign() function.
; Syntax:			_HotKeyRelease (  )
; Parameter(s):		None.
; Return Value(s):	None.
; Author(s):		Yashied
; Note(s):			-
;====================================================================================================================================

func _HotKeyRelease()
	$hkTb[0][1] = 1
	$hkTb[0][0] = 0
	redim $hkTb[1][9]
	if _WinAPI_UnhookWindowsHookEx($hkTb[0][4]) then
		$hkTb[0][4] = 0
	endif
	$hkTb[0][2] = 0
	$hkTb[0][1] = 0
endfunc; _HotKeyRelease

#EndRegion Public Functions

#Region Internal Functions

func _hold($vkCode)
	
	local $ret = DllCall('user32.dll', 'int', 'GetAsyncKeyState', 'int', $vkCode)
	
	if (not @error) and (BitAND($ret[0], 0x8000) = 0x8000) then
		return 1
	endif
	return 0
endfunc; _hold

func _hook($iCode, $wParam, $lParam)
	if $iCode > -1 then

		if $hkTb[0][1] = 1 then
			switch $wParam
				case $WM_KEYDOWN, $WM_SYSKEYDOWN
					return -1
				case else
					return _WinAPI_CallNextHookEx($hkTb[0][4], $iCode, $wParam, $lParam)
			endswitch
		endif

		switch $wParam
			case $WM_KEYDOWN, $WM_SYSKEYDOWN

				local $vkCode = DllStructGetData(DllStructCreate($tagKBDLLHOOKSTRUCT, $lParam), 'vkCode')
				local $i, $j, $call = false

				for $i = 1 to $hkTb[0][0]
					if _winactive($hkTb[$i][7]) then
						if $vkCode = BitAND($hkTb[$i][0], 0x00FF) then 
							if BitShift($hkTb[$i][0], 8) = BitOR(_hold(0x10), BitShift(_hold(0x11), -1), BitShift(_hold(0x12), -2), BitShift(_hold(0x5B) or _hold(0x5C), -3)) then
								if $hkTb[0][2] = $hkTb[$i][0] then
									if $hkTb[$i][2] > 0 then
										if TimerDiff($hkTb[$i][4]) > $hkTb[$i][2] then
											if ($hkTb[$i][5] = 0) or (TimerDiff($hkTb[$i][5]) > $hkTb[$i][3]) then
												$hkTb[$i][5] = TimerInit()
												$call = true
											endif
										endif
									endif
								else
									for $j = 1 to $hkTb[0][0]
										if $j = $i then
											if $hkTb[$i][2] > 0 then
												$hkTb[$i][4] = TimerInit()
											endif
											continueloop
										endif
										$hkTb[$j][4] = 0
										$hkTb[$j][5] = 0
									next
									$hkTb[0][2] = $hkTb[$i][0]
									$call = true
								endif
								if $call then
									Call($hkTb[$i][1])
;									if (@error = 0xDEAD) and (@extended = 0xBEEF) then
;										$hkTb[0][0] = 0
;										$hkTb[0][3] = 0
;										_WinAPI_ShowError($hkTb[$i][1] & '(): Function does not exist or invalid number of parameters.')
;										exit
;									endif
								endif
								if $hkTb[$i][6] = 1 then
									exitloop
								endif
								return -1
							endif
						endif
					endif
				next
			case else
				$hkTb[0][2] = 0
		endswitch
	endif

	return _WinAPI_CallNextHookEx($hkTb[0][4], $iCode, $wParam, $lParam)
endfunc; _hook

func _winactive($iTitle)
	if (IsNumber($iTitle)) and ($iTitle = 0) then
		return 1
	else
		if WinActive($iTitle) then
			return 1
		endif
	endif
	return 0
endfunc; _winactive

func OnHotKeyExit()
	_WinAPI_UnhookWindowsHookEx($hkTb[0][4])
	Call($OnHotKeyExit)
endfunc; OnHotKeyExit

#EndRegion Internal Functions
